@@ -14,18 +14,18 @@ import struct
 import sys
 from time import sleep, time
 
-import cbor
 import click
 
 if "linux" in platform.platform().lower():
     import fcntl
 
 # @fixme: 1st layer `nkfido2` lower layer `fido2` not to be used here !
+from fido2.cbor import dump_dict
 from fido2.client import ClientError as Fido2ClientError
 from fido2.ctap import CtapError
 from fido2.ctap1 import ApduError
-from fido2.ctap2 import Ctap2
-from fido2.ctap2.pin import ClientPin
+from fido2.ctap2 import CredentialManagement, Ctap2
+from fido2.ctap2.pin import ClientPin, PinProtocol
 
 import pynitrokey
 import pynitrokey.fido2 as nkfido2
@@ -33,18 +33,27 @@ import pynitrokey.fido2.operations
 from pynitrokey.cli.monitor import monitor
 from pynitrokey.cli.program import program
 from pynitrokey.cli.update import update
+from pynitrokey.fido2 import client
+from pynitrokey.fido2.client import NKFido2Client
 from pynitrokey.fido2.commands import SoloBootloader
-from pynitrokey.helpers import AskUser, local_critical, local_print
+from pynitrokey.helpers import (
+    AskUser,
+    local_critical,
+    local_print,
+    require_windows_admin,
+)
 
 # @todo: in version 0.4 UDP & anything earlier inside fido2.__init__ is broken/removed
 #        - check if/what is needed here
 #        - revive UDP support
 
 # https://pocoo-click.readthedocs.io/en/latest/commands/#nested-handling-and-contexts
+
+
 @click.group()
 def fido2():
     """Interact with Nitrokey FIDO2 devices, see subcommands."""
-    pass
+    require_windows_admin()
 
 
 @click.group()
@@ -184,6 +193,98 @@ def list():
         local_print(f"{id_}: {name}")
 
 
+@click.command()
+@click.option(
+    "-s",
+    "--serial",
+    help="Serial number of Nitrokey to use. Prefix with 'device=' to provide device file, e.g. 'device=/dev/hidraw5'.",
+)
+@click.option("--pin", help="provide PIN instead of asking the user", default=None)
+def list_credentials(serial, pin):
+    """List all credentials saved on the key as well as the amount of remaining slots."""
+
+    # Makes sure pin exists
+    if not pin:
+        pin = AskUser.hidden("Please provide pin: ")
+
+    nk_client = NKFido2Client()
+    cred_manager = nk_client.cred_mgmt(serial, pin)
+
+    # Returns Sequence[Mapping[int, Any]]
+    # Use this to get all existing creds
+    cred_metadata = cred_manager.get_metadata()
+    cred_count = cred_metadata.get(CredentialManagement.RESULT.EXISTING_CRED_COUNT)
+    remaining_cred_space = cred_metadata.get(
+        CredentialManagement.RESULT.MAX_REMAINING_COUNT
+    )
+
+    reliable_party_list = cred_manager.enumerate_rps()
+
+    if cred_count == 0:
+        local_print("There are no registered credentials")
+        local_print(
+            f"There is an estimated amount of {remaining_cred_space} credential slots left"
+        )
+        return
+
+    # Get amount of registered creds from first key in list (Same trick is used in the CredentialManager)
+    local_print(f"There are {cred_count} registered credentials")
+
+    for reliable_party_result in reliable_party_list:
+        reliable_party = reliable_party_result.get(CredentialManagement.RESULT.RP)
+        reliable_party_hash = reliable_party_result.get(
+            CredentialManagement.RESULT.RP_ID_HASH
+        )
+        local_print("-----------------------------------")
+        local_print(f"{reliable_party['name']}: ")
+        for cred in cred_manager.enumerate_creds(reliable_party_hash):
+            cred_id = cred.get(CredentialManagement.RESULT.CREDENTIAL_ID)["id"]
+            cred_user = cred.get(CredentialManagement.RESULT.USER)
+            if cred_user["name"] == cred_user["displayName"]:
+                local_print(f"- id: {cred_id.hex()}")
+                local_print(f"user: {cred_user['name']}\n")
+            else:
+                local_print(f"- id: {cred_id.hex()}")
+                local_print(f"user: {cred_user['displayName']} ({cred_user['name']})\n")
+
+        local_print("-----------------------------------")
+    local_print(
+        f"There is an estimated amount of {remaining_cred_space} credential slots left"
+    )
+
+
+@click.command()
+@click.option(
+    "-s",
+    "--serial",
+    help="Serial number of Nitrokey to use. Prefix with 'device=' to provide device file, e.g. 'device=/dev/hidraw5'.",
+)
+@click.option("--pin", help="provide PIN instead of asking the user", default=None)
+@click.option(
+    "-cid", "--cred-id", help="Credential id of there Credential to be deleted"
+)
+def delete_credential(serial, pin, cred_id):
+    """Delete a specific credential from the key"""
+
+    if not cred_id:
+        cred_id = AskUser.hidden("Please provide credential-id")
+
+    if not pin:
+        pin = AskUser.hidden("Please provide pin: ")
+
+    nk_client = NKFido2Client()
+    cred_manager = nk_client.cred_mgmt(serial, pin)
+
+    tmp_cred_id = {"id": bytes.fromhex(cred_id), "type": "public-key"}
+
+    try:
+        cred_manager.delete_cred(tmp_cred_id)
+    except Exception as e:
+        local_critical("Failed to delete credential, was the right cred_id given?")
+        return
+    local_print("Credential was successfully deleted")
+
+
 @click.command()
 @click.option("--count", default=8, help="How many bytes to generate (defaults to 8)")
 @click.option(
@@ -317,7 +418,7 @@ def feedkernel(count, serial):
     default="Touch your authenticator to generate a credential...",
     show_default=True,
 )
-def make_credential(serial, host, user, udp, prompt, pin):
+def make_credential(serial, host, user, udp, prompt):
     """Generate a credential.
 
     Pass `--prompt ""` to output only the `credential_id` as hex.
@@ -331,9 +432,7 @@ def make_credential(serial, host, user, udp, prompt, pin):
         user_id=user,
         serial=serial,
         output=True,
-        prompt=prompt,
         udp=udp,
-        pin=pin,
     )
 
 
@@ -389,9 +488,9 @@ def challenge_response(serial, host, user, prompt, credential_id, challenge, udp
 
 
 ######
-###### @fixme: - excluded 'probe' for now, as command:
-######           SoloBootloader.HIDCommandProbe => 0x70 returns "INVALID_COMMAND"
-######         - decide its future asap...
+# @fixme: - excluded 'probe' for now, as command:
+# SoloBootloader.HIDCommandProbe => 0x70 returns "INVALID_COMMAND"
+# - decide its future asap...
 @click.command()
 @click.option(
     "-s",
@@ -431,7 +530,7 @@ def probe(serial, udp, hash_type, filename):
 
     p = nkfido2.find(serial, udp=udp)
 
-    serialized_command = cbor.dumps({"subcommand": hash_type, "data": data})
+    serialized_command = dump_dict({"subcommand": hash_type, "data": data})
     result = p.send_data_hid(SoloBootloader.HIDCommandProbe, serialized_command)
     result_hex = result.hex()
     local_print(result_hex)
@@ -560,7 +659,6 @@ def set_pin(serial):
 
 
 @click.command()
-# @click.option("--pin", help="PIN for to access key", default=None)
 @click.option(
     "-s",
     "--serial",
@@ -569,18 +667,15 @@ def set_pin(serial):
 @click.option(
     "--udp", is_flag=True, default=False, help="Communicate over UDP with software key"
 )
-def verify(serial, udp):
+@click.option("--pin", help="PIN for device access", default=None)
+def verify(serial, udp, pin):
     """Verify if connected Nitrokey FIDO2 device is genuine."""
 
-    # if not pin:
-    #    pin = AskUser("PIN required: ", repeat=0, hide_input=True).ask()
-
-    # Any longer and this needs to go in a submodule
-    local_print("please press the button on your Nitrokey device")
-
     cert = None
     try:
-        cert = nkfido2.find(serial, udp=udp).make_credential(fingerprint_only=True)
+        cert = nkfido2.find(serial, udp=udp, pin=pin).make_credential(
+            fingerprint_only=True
+        )
 
     except Fido2ClientError as e:
         cause = str(e.cause)
@@ -632,6 +727,8 @@ def verify(serial, udp):
         "e1f40563be291c30bc3cc381a7ef46b89ef972bdb048b716b0a888043cf9072a": "Nitrokey FIDO2 Dev 2.x ",
         "ad8fd1d16f59104b9e06ef323cc03f777ed5303cd421a101c9cb00bb3fdf722d": "Nitrokey 3",
         "44fa598fdc98681dc5c8659a804c40bd6e53f8e54a781608b0651d47a53e1c8a": "Nitrokey 3 Dev",
+        "aa1cb760c2879530e7d7fed3da75345d25774be9cfdbbcbd36fdee767025f34b": "Nitrokey 3 A NFC",
+        "4c331d7af869fd1d8217198b917a33d1fa503e9778da7638504a64a438661ae0": "Nitrokey 3 A Mini",
     }
 
     a_hex = cert
@@ -721,6 +818,9 @@ fido2.add_command(rng)
 fido2.add_command(reboot)
 fido2.add_command(list)
 
+fido2.add_command(list_credentials)
+fido2.add_command(delete_credential)
+
 rng.add_command(hexbytes)
 rng.add_command(raw)
 rng.add_command(feedkernel)
@@ -748,7 +848,6 @@ util.add_command(genkey)
 util.add_command(mergehex)
 util.add_command(monitor)
 
-
 # see above -> @fixme: likely to be removed?!
 # fido2.add_command(probe)
 # key.add_command(sha256sum)
